from django.apps import AppConfig


class StoreConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'hostore'
    verbose = 'Holmium Load Curve Store'
    label = 'hostore'



import datetime as dt
import logging
import pandas as pd
from typing import Union, List, Tuple

logger = logging.getLogger(__name__)


def check_ts_completeness(ds: pd.Series, start: dt.datetime, end: dt.datetime, tag=None, freq=None, freq_margin=None,
                          msgs=None) -> List[Tuple[Union[dt.datetime, dt.date], Union[dt.datetime, dt.date]]]:
    """
    Check if timeserie ds contains full data between start and end at freq.

    Optionnally store descriptions of holes in msgs.

    :param ds: timeserie to check
    :type ds: pd.DataFrame or pd.Series
    :param start: check start
    :type start: dt.datetime
    :param end: check end
    :type end: dt.datetime
    :param tag: tag of ds for logging purpose
    :type tag: None/str
    :param freq: freq to reindex input, default 30min
    :type freq: None/str
    :param freq_margin: freq of external data store. This is used to add margin around holes datetime ranges.
    :type freq_margin: None/str
    :param msgs: if specified, warning messages will be stored inside
    :type msgs: None/list
    :return: list of tuple [(start_hole1, end_hole1), (start_hole2, end_hole2)..]
    :rtype: list
    """
    def _notify(msg):
        fullmsg = tag + msg
        if msgs is not None:
            msgs.append(fullmsg)
        else:
            logger.warning(fullmsg)

    freq = freq if freq is not None else '30min'
    freq_margin = freq_margin if freq_margin is not None else freq
    if isinstance(freq_margin, pd.Timedelta):
        timedelta_margin = freq_margin
    else:
        timedelta_margin = pd.to_timedelta(pd_freq_fix(freq_margin))
    tag = tag + ' ' if tag is not None else ''
    if ds.empty:
        _notify(f'is empty')
        return (start, end),

    ds = ds.sum(axis=1) if isinstance(ds, pd.DataFrame) else ds
    ds_ri = ds.copy().reindex(pd.date_range(start, end, freq=freq))
    ds_nulls = ds_ri.isnull()

    nulls_seqs = []
    if ds_nulls.any():
        _notify('contains NaNs')
        seqs = find_constant_sequences(ds_nulls)
        for (_from, _to), isnull in seqs:
            if isnull:
                _notify(f'  from {_from} to {_to} ({_to-_from}) (holes widen by {timedelta_margin})')
                nulls_seqs.append((_from - timedelta_margin, _to + timedelta_margin))
    return nulls_seqs


def equalp(x, y):
    """
    Equivalent of x == y, but return True for nan == nan
    """
    return (x == y) or (pd.isna(x) and pd.isna(y))


def find_constant_sequences(ds: pd.Series):
    """
    return sequences where data is constant

    :param ds: serie to check
    :type ds: pd.Series
    :return:(([idxmin1, idxmax1], value1), ([idxmin2, idxmax2], value2), ...)
    :rtype: tuple
    """
    store = []
    any_nan = ds.isnull().any()
    for index, value in ds.items():
        if not store:
            # init
            same = False
        elif any_nan:
            # slower algo because nan == nan returns False
            same = equalp(value, store[-1][1])
        else:
            same = value == store[-1][1]

        if same:
            # edit end portion
            store[-1][0][1] = index
        else:
            # init portion
            store.append(([index, index], value))
    return tuple((tuple(x), y) for x, y in store)


def split_ts(ds: pd.Series, idx_split: dt.datetime, idx_min=None, idx_max=None):
    """
    split ts in 2 ts before and after index

    :param ds: timeserie to split
    :type ds: pd.Series
    :param idx_split: index to split timeserie
    :type idx_split: dt.datetime
    :param idx_min: minimum index (must be lower than idx_split)
    :type idx_max: dt.datetime / None
    :param idx_max: maximum index (must be greater than idx_split)
    :type idx_max: dt.datetime / None
    :return: ds_before_idx, ds_after_idx
    :rtype:
    """
    if idx_min is None:
        ds_before = ds[:idx_split]
    else:
        if idx_min >= idx_split:
            raise ValueError(f'idx_min >= idx_split : {idx_min} <= {idx_split}')
        ds_before = ds[idx_min:idx_split]

    if idx_max is None:
        ds_after = ds[idx_split:]
    else:
        if idx_max <= idx_split:
            raise ValueError(f'idx_max <= idx_split : {idx_max} <= {idx_split}')
        ds_after = ds[idx_split:idx_max]
    return ds_before, ds_after


def ts_combine_first(ds_list, default_type=None):
    """
    combine time series list : apply combine_first from first list, handle empty series

    :param ds_list: list of series to combine
    :type ds_list: [pd.Series]
    :param default_type: specify default returned empty object : Series / DataFrame
    :type default_type: callable
    :return: combined serie
    :rtype: pd.Series
    """
    ds_combined = None
    default_type = default_type if default_type is not None else pd.Series
    for ds in ds_list:
        if ds is None or ds.empty:
            continue
        if ds_combined is None:
            ds_combined = ds
        else:
            try:
                ds_combined = ds_combined.combine_first(ds)
            except Exception as ex:
                logger.error(f'ts_combine_first failed')
                logger.error(f'left serie = \n{ds_combined}')
                logger.error(f'right serie = \n{ds}')
                raise ex

    return ds_combined if ds_combined is not None else default_type(dtype=float)


def upsample_using_reference_serie(ds_resampled: pd.Series, ds_reference: pd.Series):
    """
    compute dataserie from ds_resampled data (great period) using evolutions from ds_reference (small period)

    :param ds_resampled: serie to resample, must have a greater frequency than ds_reference
    :type ds_resampled: pd.Serie
    :param ds_reference: serie with lower frequency, used to compute mapping keys
    :type ds_reference: pd.Serie
    todo : robust input checks
    todo ? exploiter l'evolution intrinseque de ds_resampled ? typiquement si un nuage passe le matin en phase de montée
    :return:
    :rtype:
    """
    add_last = True
    df_work = pd.DataFrame(index=ds_reference.index)
    df_work['feat'] = ds_resampled
    df_work['idx_time'] = pd.Series(index=ds_resampled.index, data=range(0, len(ds_resampled)))
    df_work['feat'] = df_work['feat'].ffill()
    df_work['idx_time'] = df_work['idx_time'].ffill()
    df_work['ref'] = ds_reference

    ds_out = pd.Series(index=ds_reference.index, dtype=float)
    # if debug:
    #     ds_ratio_plot = pd.Series(index=ds_reference.index, data=1.)

    df_grouped = list(df_work.groupby('idx_time'))
    if add_last:
        df_grouped_new = list()
        for ii, (idx_time, df_group) in enumerate(df_grouped):
            try:
                df_next = df_grouped[ii+1][1]
            except IndexError:
                df_next = None
            if df_next is not None:
                # df_new = df_group.append(df_next.iloc[0, :])
                added_serie = df_next.iloc[0, :]
                df_new = pd.concat([df_group, added_serie.to_frame().transpose()])
            else:
                df_new = df_group
            df_grouped_new.append((idx_time, df_new))
        df_grouped = df_grouped_new

    for ii, (idx_time, df_group) in enumerate(df_grouped):
        ref_mean = df_group['ref'].mean()
        if ref_mean != 0:
            ds_ratio = df_group['ref'] / ref_mean
            ds_out[df_group.index] = ds_ratio * df_group['feat']
            # if debug:
            #     ds_ratio_plot[df_group.index] = ds_ratio
        else:
            ds_out[df_group.index] = df_group['feat']

    # if debug:
    #     plotter = Plotter()
    #     plotter.plot_feat(df_work['feat'], label='feat', marker='x')
    #     plotter.plot_feat(df_work['ref'], label='ref', marker='x')
    #     plotter.plot_feat(ds_out, label='feat reasampled', marker='x')
    #     # plotter.plot_feat(ds_ratio_plot, label='ratios computed', marker='x', secondary_axis=True)
    #     plotter.show_plot('upsample_using_reference_serie')
    return ds_out


def pd_freq_fix(freq_str: str):
    """
    Fix the syntax of frequency returned by some pandas methods.
    For instance :
        In[]:  pd.infer_freq(ds.index)
        Out[]: 'H'
        pd.to_timedelta('H') will raise a ValueError
    :param freq_str:
    :type freq_str:
    :return: fixed frequency syntax
    :rtype: str
    """
    if not isinstance(freq_str, str):
        logger.warning(f'freq_str {freq_str} is not a string')
        return freq_str
    elif freq_str[0].isnumeric():
        return freq_str
    else:
        return '1' + freq_str


def pd_min_freq(index):
    """
    Find minimum frequency in index
    :param index:
    :return:
    """
    if len(index) < 2:
        return None

    diffs = (index[1:] - index[:-1])
    min_delta = diffs.min()
    return abs(min_delta)


def ts_normalize(ds: pd.Series):
    """ normalize series to be in [0, 1] """
    # (ds - ds.mean()) / ds.std()
    return (ds - ds.min()) / (ds.max() - ds.min())


def slice_ts(ds: Union[pd.Series, pd.DataFrame], start: dt.datetime, end: dt.datetime):
    pos = ((ds.index >= start) &
           (ds.index < end))
    return ds.loc[pos]



# Generated by Django 5.0.2 on 2024-03-21 16:21

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('hostore', '0001_initial'),
    ]

    operations = [
        migrations.RenameIndex(
            model_name='testdatastore',
            new_name='hostore_tes_prm_897468_idx',
            old_name='store_testd_prm_91e86e_idx',
        ),
    ]



import io
import logging
from collections import defaultdict
from typing import List, Dict
import datetime as dt
import pandas as pd
from django.db import models
from django.db.models import Max
from pyarrow import BufferReader

from .manager import StoreQuerySet
from .utils.timeseries import ts_combine_first, find_constant_sequences, check_ts_completeness
from .utils.utils import chunks
from django.utils.timezone import now

logger = logging.getLogger(__name__)


class Store(models.Model):
    client_id = models.IntegerField()
    prm = models.CharField(max_length=30)
    last_modified = models.DateTimeField(auto_now=True)
    created_at = models.DateTimeField(auto_now_add=True, null=True)
    data = models.BinaryField(blank=True, null=True)
    version = models.IntegerField(default=0)

    objects = StoreQuerySet.as_manager()

    class Meta:
        abstract = True
        unique_together = ('prm', 'client_id', 'version')
        indexes = [models.Index(fields=['prm', 'client_id']), ]

    @classmethod
    def count(cls, client_id: int):
        return cls.objects.filter(client_id=client_id).count()

    @classmethod
    def find_holes(cls, client_id: int, sd: dt.datetime, ed: dt.datetime, freq='30min', prms: List[str] = None, chunk_size=50,
                   freq_margin=pd.Timedelta(minutes=0)) -> Dict[str, List]:
        """
        Find holes (missing data periods) for multiple prms within a specified datetime range.

        This method retrieves data from a database for each parameter associated with a client_id,
        checks each dataset for missing data within the specified datetime range, and identifies the start
        and end of each missing data period.

        :param client_id: ID of the client for whom the data is being checked.
        :param sd: Start datetime of the range to check for holes.
        :param ed: End datetime of the range to check for holes.
        :param freq: Frequency at which the data is expected, defaulting to '30min'.
        :param prms: Optional list of prms to check; if None, all parameters for the client are checked.
        :param chunk_size: The size of chunks to break the parameter list into for batch processing.
                           This helps in managing memory and processing large datasets efficiently, defaulting to 50.
        :param freq_margin: Freq margin for null sequences
        :return: A dictionary with prm as keys and a list of tuples (start, end) indicating
                 the missing data periods for each parameter. Start and end are included
        """
        # If no prm are specified, retrieve all prm for the given client from the database
        if prms is None:
            prms = list(cls.objects.filter(client_id=client_id).values_list('prm', flat=True))

        # Iterate over prms in chunks to manage memory usage and efficiency
        for prm_chunk in chunks(prms, chunk_size):
            data_by_prm = cls.get_many_lc(prm_chunk, client_id, combined_versions=True)

            # Check each parameter's data for completeness
            for prm in prm_chunk:
                nulls_seqs = []
                data = data_by_prm[prm]
                if len(data) == 0:
                    # If there's no data for a prm, the entire range is considered as a hole
                    nulls_seqs = [(sd, ed)]
                else:
                    # Data is wrapped in a list, and there's only one dataset per prm (combined_versions=True)
                    assert len(data) == 1
                    nulls_seqs = check_ts_completeness(data[0]['data'], sd, ed, freq=freq, freq_margin=freq_margin)

                # Yield the output for this prm
                # This allows processing part by part without waiting for the entire operation to complete
                yield prm, nulls_seqs

    @classmethod
    def get_lc(cls, prm: str, client_id: int, combined_versions=True, version: int = None) -> List[Dict]:
        """
        Get the prm load curve
        Args:
            prm: str
            client_id: int, client's id
            combined_versions: bool, if multiple versions exists we combine them (default True)
            version: return only the selected version
        Returns:
            List[Dict]
        """
        logger.info(f'Getting load curve with prm {prm} for client {client_id}')
        qs = cls.objects.filter(prm=prm, client_id=client_id).order_by('-version')
        if version is not None:
            qs = qs.filter(version=version)

        entries = []
        for entry in qs:
            entry_dict = vars(entry)
            reader = BufferReader(entry_dict['data'])
            ds = pd.read_feather(reader)
            if 'index' in ds.columns:
                ds.set_index('index', inplace=True)
            else:
                logger.info(f'data is malformed, remove key {prm}')
                entry.delete()
            entry_dict['data'] = ds.iloc[:, 0]
            entries.append(entry_dict)

        if len(entries) > 0 and combined_versions:
            ds_combined = ts_combine_first([e['data'] for e in entries])
            entries[0]['data'] = ds_combined
            return entries[0:1]
        return entries

    @classmethod
    def get_many_lc(cls, prms: [str], client_id: int, combined_versions=True) -> Dict[str, List[Dict]]:
        """
        Get many prms from cache
        Args:
            prms: list of prms
            client_id: int, client's id
            combined_versions: bool, if multiple versions exists we combine them (default True)
        Returns:
            Dict[str, List[Dict]]
        """
        qs = cls.objects.filter(prm__in=prms, client_id=client_id).order_by('-version')
        results = defaultdict(lambda: [])
        invalid_ids = []
        for entry in qs:
            reader = BufferReader(entry.data)
            ds = pd.read_feather(reader)
            try:
                ds.set_index('index', inplace=True)
            except KeyError:
                invalid_ids.append(entry.id)
                continue
            entry_dict = vars(entry)
            entry_dict['data'] = ds.iloc[:, 0]
            results[entry.prm].append(entry_dict)

        if len(invalid_ids) > 0:
            logger.info(f'data is malformed, for ids {invalid_ids}')
            cls.objects.filter(id__in=invalid_ids).delete()

        if len(results) > 0 and combined_versions:
            for prm, entries in results.items():
                ds_combined = ts_combine_first([e['data'] for e in entries])
                entries[0]['data'] = ds_combined
                results[prm] = entries[0:1]

        return results

    @classmethod
    def set_lc(cls, prm: str, value: pd.Series, client_id: int, versionning=False) -> None:
        """
        Set one prm to cache
        Args:
            prm: str
            value: pd.Series
            client_id: int, client's id
            versionning: bool, if True, a new version will be saved else the load curve is replaced (default False)

        Returns:
            None
        """
        logger.info(f'SET key {prm} in cache')
        if isinstance(value, pd.Series):
            if value.isnull().all():
                logger.warning(f'CACHE : Key {prm} is ignored because data is null')
                return
            df = value.to_frame(name=prm)
            df.reset_index(inplace=True, names=['index'])
            buf = io.BytesIO()
            df.to_feather(buf, compression='lz4')
            v = buf.getvalue()
            if not versionning:
                cls.objects.update_or_create(client_id=client_id, prm=prm, defaults=dict(data=v))
            else:
                latest_version = cls.objects.filter(client_id=client_id, prm=prm).aggregate(Max('version'))[
                    'version__max']
                if latest_version is not None:
                    # If there's at least one object with the same prm, increment the version
                    version = latest_version + 1
                else:
                    # If there are no objects with the same prm, start with version 0
                    version = 0
                logger.info(f"version to insert {version}")
                cls.objects.create(client_id=client_id, prm=prm, data=v, version=version)
            logger.info(f'SET prm {prm} in cache DONE')
        else:
            raise ValueError(f'Cannot cache value of type {type(value)}, only dataframe are accepted')

    @classmethod
    def set_many_lc(cls, dataseries: Dict[str, pd.Series], client_id: int, versionning=False) -> None:
        """
        Update prms contained in dataseries dict
        Args:
            dataseries: dict(key: pd.Series)
            client_id: int, client's id
            versionning: bool, if True, a new version will be saved for each prm else the load curve is replaced
        Returns:
            bool
        """
        if len(dataseries) > 0:
            logger.info(f'Updating cache for {list(dataseries.keys())}')
        for prm, data in dataseries.items():
            if data is not None and not data.empty:
                cls.set_lc(prm, data, client_id, versionning=versionning)

    @classmethod
    def clear(cls, prms: [str], client_id: int, version=None) -> None:
        """
        Method to clear multiple prm from cache
        Args:
            prms: list of prms
            client_id: client's id
            version: version id to delete
        Returns:
            None
        """
        qs = cls.objects.filter(prm__in=prms, client_id=client_id)
        if version is not None:
            qs = qs.filter(version=version)
        qs.delete()

    @classmethod
    def clear_all(cls, client_id: int = None) -> None:
        """
        Method to clear multiple prm from cache
        Args:
            client_id: client's id
        Returns:
            None
        """
        if client_id is not None:
            cls.objects.filter(client_id=client_id).delete()
        else:
            cls.objects.all().delete()


class TestDataStore(Store):
    class Meta(Store.Meta):
        abstract = False
        app_label = 'hostore'



# Generated by Django 5.0.2 on 2024-02-21 11:40

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='TestDataStore',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('client_id', models.IntegerField()),
                ('prm', models.CharField(max_length=30)),
                ('last_modified', models.DateTimeField(auto_now=True)),
                ('data', models.BinaryField(blank=True, null=True)),
            ],
            options={
                'abstract': False,
                'indexes': [models.Index(fields=['prm', 'client_id'], name='store_testd_prm_91e86e_idx')],
                'unique_together': {('prm', 'client_id')},
            },
        ),
    ]



import io
import random
import datetime as dt
import numpy as np
import pandas as pd
import pytz
from django.test import TestCase

from .models import TestDataStore


class HoCacheTestCase(TestCase):
    databases = ('default',)
    def setUp(self):
        # Set up data for the tests
        # For example, create a HoCache instance
        self.test_prm = 'test_prm'
        self.test_prm_2 = 'test_prm_2'
        self.test_client_id = 1
        self.test_data = pd.Series([1, 2, 3], index=[0, 1, 2], name='data')
        self.test_data_2 = self.test_data * 2

        # Convert test data to a binary format
        df = self.test_data.to_frame(name=self.test_prm)
        df.reset_index(inplace=True, names=['index'])
        buf = io.BytesIO()
        df.to_feather(buf, compression='lz4')
        self.test_binary_data = buf.getvalue()

        TestDataStore.objects.create(
            prm=self.test_prm,
            client_id=self.test_client_id,
            data=self.test_binary_data
        )

        df = self.test_data_2.to_frame(name=self.test_prm)
        df.reset_index(inplace=True, names=['index'])
        buf = io.BytesIO()
        df.to_feather(buf, compression='lz4')
        self.test_binary_data_2 = buf.getvalue()

        TestDataStore.objects.create(
            prm=self.test_prm_2,
            client_id=self.test_client_id,
            data=self.test_binary_data,
            version=0
        )
        TestDataStore.objects.create(
            prm=self.test_prm_2,
            client_id=self.test_client_id,
            data=self.test_binary_data_2,
            version=1
        )

    def test_get_lc(self):
        # Test the get_lc method
        cache_entry = TestDataStore.get_lc(prm=self.test_prm, client_id=self.test_client_id)
        self.assertIsNotNone(cache_entry)
        self.assertEquals(len(cache_entry), 1)
        self.assertEqual(cache_entry[0]['prm'], self.test_prm)
        pd.testing.assert_series_equal(cache_entry[0]['data'], self.test_data, check_names=False)

    def test_get_many_lc(self):
        # Test the get_many_lc method
        prms = [self.test_prm, 'non_existing_prm']
        results = TestDataStore.get_many_lc(prms=prms, client_id=self.test_client_id)
        self.assertIn(self.test_prm, results)
        self.assertNotIn('non_existing_prm', results)
        self.assertEqual(len(results[self.test_prm]), 1)
        pd.testing.assert_series_equal(results[self.test_prm][0]['data'], self.test_data, check_names=False)

    def test_set_many_lc(self):
        # Test the get_many_lc method
        dataseries = dict(prm_1=pd.Series([1, 2, 3]),
                          prm_2=pd.Series([4, 5, 6]))
        TestDataStore.set_many_lc(dataseries, self.test_client_id)
        results = TestDataStore.get_many_lc(['prm_1', 'prm_2'], self.test_client_id)
        self.assertIn("prm_1", results)
        self.assertIn("prm_2", results)
        self.assertEqual(len(results), 2)
        self.assertNotIn('non_existing_prm', results)
        self.assertEqual(len(results['prm_1']), 1)
        self.assertEqual(len(results['prm_2']), 1)
        pd.testing.assert_series_equal(results['prm_1'][0]['data'], dataseries['prm_1'], check_names=False)
        pd.testing.assert_series_equal(results['prm_2'][0]['data'], dataseries['prm_2'], check_names=False)

    def test_set_lc(self):
        # Test the set_lc method
        new_prm = 'new_test_prm'
        TestDataStore.set_lc(prm=new_prm, value=self.test_data, client_id=self.test_client_id)
        # We should find the prm
        TestDataStore.objects.get(prm=new_prm, client_id=self.test_client_id, version=0)

        # Récupération des données du PRM sans spécifier de version
        data = TestDataStore.get_lc(new_prm, self.test_client_id)
        self.assertEquals(len(data), 1)
        pd.testing.assert_series_equal(data[0]['data'], self.test_data, check_names=False)

        # Récupération des données du PRM en spécifiant la version
        data = TestDataStore.get_lc(new_prm, self.test_client_id, version=0)
        self.assertEquals(len(data), 1)
        pd.testing.assert_series_equal(data[0]['data'], self.test_data, check_names=False)

    def test_set_existing_lc(self):
        # Test the set_lc method
        new_prm = 'new_test_param'
        # Insert 2 times in a row without versionning
        TestDataStore.set_lc(prm=new_prm, value=self.test_data, client_id=self.test_client_id, versionning=False)
        TestDataStore.set_lc(prm=new_prm, value=self.test_data, client_id=self.test_client_id, versionning=False)

        data = TestDataStore.get_lc(new_prm, self.test_client_id, version=0)
        self.assertEquals(len(data), 1)
        self.assertEquals(data[0]['version'], 0)

        # Insert 2 times in a row with versionning
        TestDataStore.set_lc(prm=new_prm, value=self.test_data, client_id=self.test_client_id, versionning=True)
        TestDataStore.set_lc(prm=new_prm, value=self.test_data, client_id=self.test_client_id, versionning=True)

        data = TestDataStore.get_lc(new_prm, self.test_client_id, combined_versions=False)
        self.assertEquals(len(data), 3)
        self.assertEquals(data[0]['version'], 2)

        data = TestDataStore.get_lc(new_prm, self.test_client_id, combined_versions=True)
        self.assertEquals(len(data), 1)
        self.assertEquals(data[0]['version'], 2)

    def test_combined_versions(self):
        data = TestDataStore.get_lc(self.test_prm_2, self.test_client_id, combined_versions=True)
        self.assertEquals(len(data), 1)
        self.assertEquals(data[0]['version'], 1)
        # Vérifier que la combinaison donne la deuxième version
        pd.testing.assert_series_equal(data[0]['data'], self.test_data_2, check_names=False)

        # test if we specify a version
        data = TestDataStore.get_lc(self.test_prm_2, self.test_client_id, combined_versions=False, version=1)
        self.assertEquals(len(data), 1)
        self.assertEquals(data[0]['version'], 1)
        # Vérifier que la combinaison donne la deuxième version
        pd.testing.assert_series_equal(data[0]['data'], self.test_data_2, check_names=False)

    def test_non_combined_versions(self):
        data = TestDataStore.get_lc(self.test_prm_2, self.test_client_id, combined_versions=False)
        self.assertEquals(len(data), 2)
        self.assertEquals(data[0]['version'], 1)
        self.assertEquals(data[1]['version'], 0)
        # Vérifier que la combinaison donne la deuxième version
        pd.testing.assert_series_equal(data[0]['data'], self.test_data_2, check_names=False)
        pd.testing.assert_series_equal(data[1]['data'], self.test_data, check_names=False)

        # test if we specify a version
        data = TestDataStore.get_lc(self.test_prm_2, self.test_client_id, combined_versions=False, version=0)
        self.assertEquals(len(data), 1)
        self.assertEquals(data[0]['version'], 0)
        # Vérifier que la combinaison donne la deuxième version
        pd.testing.assert_series_equal(data[0]['data'], self.test_data, check_names=False)

    # *******************************************************

    def test_clearing(self):
        TestDataStore.clear([self.test_prm], self.test_client_id)
        data_after_clear = TestDataStore.get_lc(self.test_prm, self.test_client_id)
        self.assertEqual(len(data_after_clear), 0)

    def test_clearing_version(self):
        TestDataStore.clear([self.test_prm_2], self.test_client_id, version=0)
        data_after_clear = TestDataStore.get_lc(self.test_prm, self.test_client_id, combined_versions=False)
        self.assertEqual(len(data_after_clear), 1)

    def test_clearing_all(self):
        TestDataStore.clear_all()
        data_after_clear = TestDataStore.get_lc(self.test_prm, self.test_client_id)
        self.assertEqual(len(data_after_clear), 0)

    def test_set_get_clear_dataframe(self):
        idx = pd.date_range('2022-01-01', '2024-01-01', freq='30min')
        data = [random.random() for k in range(0, len(idx))]
        ds_initial = pd.Series(data, index=idx)
        new_client_id = self.test_client_id + 1
        # Set key, value in cache
        TestDataStore.set_lc('key', ds_initial, new_client_id, versionning=False)
        # Get key, value from cache
        data = TestDataStore.get_lc('key', new_client_id)
        self.assertEqual(ds_initial.compare(data[0]['data']).shape[0], 0)
        TestDataStore.clear_all(client_id=new_client_id)
        # 3 entrées sur le client_id self.test_client_id
        self.assertEqual(TestDataStore.count(self.test_client_id), 3)
        # 3 entrées sur le client_id self.test_client_id
        self.assertEqual(TestDataStore.count(self.test_client_id + 1), 0)

    def test_set_dataframe_with_nan(self):
        idx = pd.date_range('2022-01-01', '2024-01-01', freq='30min')
        ds_initial = pd.Series(np.nan, index=idx)

        # Set key, value in cache
        TestDataStore.set_lc('key', ds_initial, self.test_client_id)
        # Get key, value from cache
        data = TestDataStore.get_lc('key', self.test_client_id)
        self.assertEqual(len(data), 0)

    def test_not_existing_key(self):
        data = TestDataStore.get_lc('not_existing_key', self.test_client_id)
        self.assertEqual(len(data), 0)

    def test_cache_not_series(self):
        self.assertRaises(ValueError, TestDataStore.set_lc, 'key', 'hello', self.test_client_id)

    def test_set_index_with_different_names(self):
        ds = pd.Series([1, 2, 3])
        ds.index.name = 'test'
        TestDataStore.set_lc('key', ds, self.test_client_id)
        # Get the serie from cache
        ds_from_cache = TestDataStore.get_lc('key', self.test_client_id)[0]['data']
        pd.testing.assert_series_equal(ds, ds_from_cache, check_names=False)

    def test_find_holes(self):
        sd = dt.datetime(2024, 1, 1).astimezone(tz=pytz.UTC)
        ed = dt.datetime(2024, 1, 11).astimezone(tz=pytz.UTC)
        idx = pd.date_range(sd, ed, freq='D')
        ds_no_hole = pd.Series(10, idx)
        ds_no_hole.index.name = 'no_hole'

        ds_one_hole = ds_no_hole.copy()
        ds_one_hole.loc['2024-01-02':'2024-01-03'] = None
        ds_one_hole.index.name = 'one_hole'

        ds_two_holes = ds_no_hole.copy()
        ds_two_holes.loc['2024-01-02':'2024-01-03'] = None
        ds_two_holes.loc['2024-01-05':'2024-01-08'] = None
        ds_two_holes.index.name = 'two_holes'

        TestDataStore.set_lc('no_hole', ds_no_hole, self.test_client_id)
        TestDataStore.set_lc('one_hole', ds_one_hole, self.test_client_id)
        TestDataStore.set_lc('two_holes', ds_two_holes, self.test_client_id)

        # Get holes of one_hole
        holes = list(TestDataStore.find_holes(self.test_client_id, sd, ed, freq='D', prms=['one_hole']))
        self.assertEquals(len(holes), 1)
        self.assertEquals(holes[0][0], 'one_hole')
        self.assertEquals(holes[0][1][0],
                          (dt.datetime(2024, 1, 2).astimezone(tz=pytz.UTC),
                           dt.datetime(2024, 1, 3).astimezone(tz=pytz.UTC)))

        # Get holes of two_holes
        holes = list(TestDataStore.find_holes(self.test_client_id, sd, ed, freq='D', prms=['two_holes']))
        self.assertEquals(len(holes), 1)
        self.assertEquals(holes[0][0], 'two_holes')
        self.assertEquals(holes[0][1][0],
                          (dt.datetime(2024, 1, 2).astimezone(tz=pytz.UTC),
                           dt.datetime(2024, 1, 3).astimezone(tz=pytz.UTC)))
        self.assertEquals(holes[0][1][1],
                          (dt.datetime(2024, 1, 5).astimezone(tz=pytz.UTC),
                           dt.datetime(2024, 1, 8).astimezone(tz=pytz.UTC)))

        # Get holes of no_hole
        holes = list(TestDataStore.find_holes(self.test_client_id, sd, ed, freq='D', prms=['no_hole']))
        self.assertEquals(holes[0][0], 'no_hole')
        self.assertEquals(holes[0][1], [])

        # Get all holes
        holes = list(TestDataStore.find_holes(self.test_client_id, sd, ed, freq='D', prms=['no_hole', 'one_hole', 'two_holes', 'non_existing']))
        self.assertEquals(len(holes), 4)



# Generated by Django 5.0.2 on 2024-05-03 13:57

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('hostore', '0003_alter_testdatastore_unique_together_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='testdatastore',
            name='created_at',
            field=models.DateTimeField(auto_now_add=True, null=True),
        ),
    ]



# Generated by Django 5.0.2 on 2024-04-03 20:09

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('hostore', '0002_rename_store_testd_prm_91e86e_idx_hostore_tes_prm_897468_idx'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='testdatastore',
            unique_together=set(),
        ),
        migrations.AddField(
            model_name='testdatastore',
            name='version',
            field=models.IntegerField(default=0),
        ),
        migrations.AlterUniqueTogether(
            name='testdatastore',
            unique_together={('prm', 'client_id', 'version')},
        ),
    ]
